'use strict'

const jsonld = require('jsonld')
const jsig = require('jsonld-signatures')
const bitcore = require('bitcore-lib')
const crypto = require('crypto')
const Web3 = require('web3')

jsig.use('jsonld', jsonld)

const SSIF_SIGNATURE_ALGORITHM = 'EcdsaKoblitzSignature2016'
const SSIF_BLOCKCHAIN_ALGORITHM = 'SSIFEthereumHash2017'
const SSIF_BLOCKCHAIN_ALGORITHM_ERC2018 = 'SsifErc780Hash2018'

const SSIF_COMMITMENT_PREDICATE = 'commitment'

var TRUSTED_KEYS = {}
var TRUSTED_CONTRACTS = []
var TRUSTED_ISSUERS = []

var ETHEREUM_PROVIDER = 'http://eth4rqsua.westeurope.cloudapp.azure.com:8545';

module.exports = {
  addTrustedContract: addTrustedContract,
  addTrustedPublicKey: addTrustedPublicKey,
  addTrustedIssuer: addTrustedIssuer,
  setEthereumProvider: setEthereumProvider,
  setErc780HdKeyHash2018verifyFunction: setErc780HdKeyHash2018verifyFunction,
  getContractAt: getContractAt, // Debug
  createAttestation: createAttestation,
  formatPublicKey: formatPublicKey,
  signECDSAAttestation: signECDSAAttestation,
  verifyECDSAAttestation: verifyECDSAAttestation, // Debug
  verifyBlockchainAttestation: verifyBlockchainAttestation, // Debug
  verifyAttestation: verifyAttestation,
  verifyAttestations: verifyAttestations,
  TRUSTED_KEYS: TRUSTED_KEYS // Debug
}

// const SSIF_CONTEXT_URL = 'http://tinyurl.com/ssifschema-1'

function addTrustedPublicKey (publicKey) {
  if (!publicKey) {
    return false
  }

  TRUSTED_KEYS[publicKey] = formatPublicKey(publicKey)

  return TRUSTED_KEYS[publicKey]
}

function addTrustedContract (address) {
  if (!address) {
    return false
  }

  TRUSTED_CONTRACTS.push(address)

  return address
}

function addTrustedIssuer(address) {
  if (!address) {
    return false
  }

  TRUSTED_ISSUERS.push(address)

  return address
}

function setEthereumProvider (provider) {
  if (!provider) {
    return false
  }

  ETHEREUM_PROVIDER = provider

  return provider
}

// Define erc780Verify function - a hook to the Erc780 class verify function
var erc780Verify;

function setErc780HdKeyHash2018verifyFunction(erc780VerifyFunction) {
  erc780Verify = erc780VerifyFunction;
}

function formatPublicKey (publicKey) {
  return {
    '@context': jsig.SECURITY_CONTEXT_URL,
    id: 'ecdsa-koblitz-pubkey:' + publicKey,
    type: 'CryptographicKey',
    publicKeyWif: publicKey
  }
}

function getTrustedPublicKey (publicKey, options = {}) {
  const key = publicKey.replace(/^ecdsa-koblitz-pubkey:/, '')
  const trustedKeys = options.trustedPublicKeys || TRUSTED_KEYS
  return trustedKeys[key]
}

function getTrustedContract (contract, options = {}) {
  const address = contract.replace(/^ethereum-contract:/, '')
  const trustedContracts = options.trustedContracts || TRUSTED_CONTRACTS

  if (trustedContracts.indexOf(address) !== -1) {
    return address
  }

  return null
}

function getTrustedIssuer(issuer, options = {}) {
  const address = issuer.replace(/^ethereum-address:/, '')
  const trustedIssuers = options.trustedIssuers || TRUSTED_ISSUERS

  if (trustedIssuers.indexOf(address) !== -1) {
    return address
  }

  return null
}


async function verifyAttestations (attestations, options = {}) {
  return Promise.all(attestations.map(function (attestation) {
    return verifyAttestation(attestation, options)
  }))
}

async function verifyAttestation (attestation, options = {}) {
   if (attestation.commitment) {
    if (attestation.commitment.type === SSIF_SIGNATURE_ALGORITHM) {
      return verifyECDSAAttestation(attestation, options)
    } else if (attestation.commitment.type === SSIF_BLOCKCHAIN_ALGORITHM) {
      return verifyBlockchainAttestation(attestation, options)
    } else if (attestation.commitment.type === SSIF_BLOCKCHAIN_ALGORITHM_ERC2018) {
      return verifyBlockchainAttestationErc780(attestation, options)
    }
  }

  throw new Error('Attestation cannot be validated')
}

async function verifyECDSAAttestation (attestation, options = {}) {
  const copy = Object.assign({}, attestation)

  const commitment = copy.commitment
  delete copy.commitment

  copy['signature'] = commitment

  const publicKey = getTrustedPublicKey(commitment.creator, {trustedPublicKeys: options.trustedPublicKeys})

  if (!publicKey) {
    throw new Error('Not created by trusted public key')
  }

  const publicKeyOwner = {
    '@context': jsig.SECURITY_CONTEXT_URL,
    publicKey: [publicKey]
  }

  if (await jsig.promises.verify(copy, Object.assign({publicKey: publicKey, publicKeyOwner: publicKeyOwner}, options))) {
    return attestation
  }

  throw new Error('Not valid')
}

/**
 * Used for SSIFEthereumHash2017
 *
 * @param {*} attestation
 * @param {*} options
 */
async function verifyBlockchainAttestation (attestation, options = {}) {
  const contractAddress = getTrustedContract(attestation.commitment.contract, options)

  if (!contractAddress) {
    throw new Error('Not on trusted contract')
  }

  const contract = await getContractAt(contractAddress)
  const statements = Object.assign({}, attestation.statements)

  const address = statements['@id'].replace(/^ecdsa-koblitz-pubkey:/, '')
  const publicKeyString = attestation.commitment['publicKey'].replace(/^(ecdsa-koblitz-pubkey:)?(0x)?/, '')

  delete statements['@id']

  statements['@context'] = attestation['@context']

  console.log(statements)

  const publicKey = bitcore.PublicKey(publicKeyString).toAddress('livenet')

  if (publicKey.toString() !== address) {
    throw new Error(`Subject is not the address of ${publicKey}`)
  }

  const compacted = await jsonld.promises.compact(statements, {})
  var promises = []

  for (let predicate in compacted) {
    promises.push(verifyBlockchainStatement(contract, publicKeyString, predicate, compacted[predicate]))
  }

  await Promise.all(promises)

  return attestation
}

/**
 * used for SSIFEthereumHash2017
 *
 * @param {*} contract
 * @param {*} publicKey
 * @param {*} predicate
 * @param {*} object
 */
async function verifyBlockchainStatement (contract, publicKey, predicate, object) {
  const key = publicKey.replace(/^ecdsa-koblitz-pubkey:/, '')
  var objectValue = object

  if (object instanceof Object && '@value' in object) {
    objectValue = object['@value']
  }

  const [subjectHash, predicateHash, objectHash] = await Promise.all([
    getPublicKeyBasedHash(key),
    getPublicKeyBasedHash(key, predicate),
    getPublicKeyBasedHash(key, objectValue)
  ])

  console.log(`Checking blockchain attestation for ${publicKey} ${predicate} ${object}`)
  console.log(`subjectHash: ${subjectHash}`)
  console.log(`predicateHash: ${predicateHash}`)
  console.log(`objectHash: ${objectHash}`)

  let result = await contract.methods.verifyAttestation(subjectHash, predicateHash, objectHash).call();

  console.log(`Result = ${result}`)

  if (result) {
    let statement = {'@id': publicKey}
    statement[predicate] = objectValue

    return statement
  }

  throw new Error(`Attestation for ${publicKey} ${predicate} ${object} invalid`)
}

/**
 * Used for SSIFEthereumErc780Hash2018
 *
 * @param {*} attestation
 * @param {*} options
 */
async function verifyBlockchainAttestationErc780(attestation, options = {}) {
  const issuer = getTrustedIssuer(attestation.commitment.issuer, options)

  if (!issuer) {
    throw new Error('Not a trusted issuer')
  }

  const statements = Object.assign({}, attestation.statements)

  const address = statements['@id'].replace(/^ecdsa-koblitz-pubkey:/, '')
  const publicKeyString = attestation.commitment['publicKey'].replace(/^(ecdsa-koblitz-pubkey:)?(0x)?/, '')
  
  delete statements['@id']

  statements['@context'] = attestation['@context']

  console.log(statements)

  const publicKey = bitcore.PublicKey(publicKeyString).toAddress('livenet')

  if (publicKey.toString() !== address) {
    throw new Error(`Subject is not the address of ${publicKey}`)
  }

  const compacted = await jsonld.promises.compact(statements, {})

  for (let predicate in compacted) {
    var object = compacted[predicate]
    var objectValue = object

    if (object instanceof Object && '@value' in object) {
      objectValue = object['@value']
    }

    // Dirty fix, just make it work for now :-(
    const originalPredicate = predicate.replace(/#/, '/');

    const result = await erc780Verify(
      issuer,
      '0x' + publicKeyString,
      originalPredicate,
      objectValue
    )

    if (!result) throw new Error(`Attestation for ${publicKeyString} ${originalPredicate} ${compacted[predicate]} invalid`);
  }

  return attestation
}


function createAttestation (subject, predicate, object) {
  let statements = {}

  if (typeof subject === 'undefined') {
    throw new Error('Need to pass a subject or statements')
  }

  if (typeof predicate === 'undefined' && typeof object === 'undefined') {
    statements = subject
  } else {
    if (subject !== null) {
      statements['@id'] = subject
    }
    statements[predicate] = object
  }

  return {
    '@context': [
      'http://schema.org/',
      'https://w3id.org/security/v1',
      'https://w3id.org/identity/v1',
      'http://tinyurl.com/ssifschema-1'
    ],
    '@type': 'Attestation',
    statements: statements
  }
}

async function signECDSAAttestation (attestation, privateKey, options = {}) {
  // -> Use privateKey + jsig to sign statements.
  // -> change 'signature' to 'commitment'
  const addressFromPrivateKey = new bitcore.PrivateKey(privateKey).toPublicKey().toAddress('livenet')

  const signed = await jsig.promises.sign(attestation, Object.assign({
    privateKeyWif: privateKey,
    algorithm: SSIF_SIGNATURE_ALGORITHM,
    creator: 'ecdsa-koblitz-pubkey:' + addressFromPrivateKey
  }, options))

  const commitment = signed.signature
  delete signed.signature

  signed[SSIF_COMMITMENT_PREDICATE] = commitment

  return signed
}

const CONTRACT_ABI = [
   {
      "constant": false,
      "inputs": [],
      "name": "destroyContract",
      "outputs": [],
      "payable": false,
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "constant": false,
      "inputs": [
        {
          "name": "_subjectHash",
          "type": "address"
        }
      ],
      "name": "removeAttestation",
      "outputs": [],
      "payable": false,
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "constant": true,
      "inputs": [
        {
          "name": "_subjectHash",
          "type": "address"
        },
        {
          "name": "_predicateHash",
          "type": "address"
        },
        {
          "name": "_objectHash",
          "type": "address"
        }
      ],
      "name": "verifyAttestation",
      "outputs": [
        {
          "name": "value",
          "type": "bool"
        }
      ],
      "payable": false,
      "stateMutability": "view",
      "type": "function"
    },
    {
      "constant": true,
      "inputs": [],
      "name": "owner",
      "outputs": [
        {
          "name": "",
          "type": "address"
        }
      ],
      "payable": false,
      "stateMutability": "view",
      "type": "function"
    },
    {
      "constant": true,
      "inputs": [
        {
          "name": "_subjectHash",
          "type": "address"
        }
      ],
      "name": "getPredicateHash",
      "outputs": [
        {
          "name": "predicateHash",
          "type": "address"
        }
      ],
      "payable": false,
      "stateMutability": "view",
      "type": "function"
    },
    {
      "constant": true,
      "inputs": [
        {
          "name": "_subjectHash",
          "type": "address"
        }
      ],
      "name": "getObjectHash",
      "outputs": [
        {
          "name": "objectHash",
          "type": "address"
        }
      ],
      "payable": false,
      "stateMutability": "view",
      "type": "function"
    },
    {
      "constant": false,
      "inputs": [
        {
          "name": "_subjectHash",
          "type": "address"
        },
        {
          "name": "_predicateHash",
          "type": "address"
        },
        {
          "name": "_objectHash",
          "type": "address"
        }
      ],
      "name": "addAttestation",
      "outputs": [],
      "payable": false,
      "stateMutability": "nonpayable",
      "type": "function"
    },
    {
      "inputs": [],
      "payable": false,
      "stateMutability": "nonpayable",
      "type": "constructor"
    }
  ]

var CONTRACT, web3

async function getWeb3 () {
  // Try to catch a lost connection -- not sure if this works as intended

  if (!web3) {
    web3 = new Web3(Web3.givenProvider || ETHEREUM_PROVIDER);
  }

  await web3.eth.net.isListening().catch(() => {
    web3 = new Web3(Web3.givenProvider || ETHEREUM_PROVIDER);
  });

  return web3
}


async function getContractAt (address) {
  address = address.replace(/^ethereum-contract:/, '')

  if (!CONTRACT) {
    let web3 = await getWeb3();
    CONTRACT = new web3.eth.Contract(CONTRACT_ABI, address);
  }

  return CONTRACT;
}


function getPublicKeyBasedHash (publicKey, data) {
  let toBeHashed = Buffer.from(publicKey.replace(/^0x/, ''), 'hex')
  if (typeof data !== 'undefined' && data !== '') {
    toBeHashed = Buffer.concat([toBeHashed, Buffer.from(JSON.stringify(data))])
  }
  return '0x' + hash160(toBeHashed).toString('hex')
}

function hash160 (buf) {
  var sha = crypto.createHash('sha256').update(buf).digest()
  return crypto.createHash('rmd160').update(sha).digest()
}


/*

  Speed up JSON LD

*/

var _defaults = {
    headers: {
        accept: 'application/ld+json, application/json'
    }
};

jsonld.documentLoaders.nodeCached = function(options) {
  options = options || {};
  var headers = buildHeaders(options.headers);
  var strictSSL = ('strictSSL' in options) ? options.strictSSL : true;
  var maxRedirects = ('maxRedirects' in options) ? options.maxRedirects : -1;
  var request = ('request' in options) ? options.request : require('request');
  var http = require('http');
  // TODO: disable cache until HTTP caching implemented
  var cache = new jsonld.DocumentCache();

  var queue = new jsonld.RequestQueue();
  if(options.usePromise) {
    return queue.wrapLoader(function(url) {
      return jsonld.promisify(loadDocument, url, []);
    });
  }

  return queue.wrapLoader(function(url, callback) {
    loadDocument(url, [], callback);
  });

  function loadDocument(url, redirects, callback) {
    if(url.indexOf('http:') !== 0 && url.indexOf('https:') !== 0) {
      return callback(new JsonLdError(
        'URL could not be dereferenced; only "http" and "https" URLs are ' +
        'supported.',
        'jsonld.InvalidUrl', {code: 'loading document failed', url: url}),
        {contextUrl: null, documentUrl: url, document: null});
    }
    if(options.secure && url.indexOf('https') !== 0) {
      return callback(new JsonLdError(
        'URL could not be dereferenced; secure mode is enabled and ' +
        'the URL\'s scheme is not "https".',
        'jsonld.InvalidUrl', {code: 'loading document failed', url: url}),
        {contextUrl: null, documentUrl: url, document: null});
    }
    // TODO: disable cache until HTTP caching implemented
    var doc = cache.get(url);
    if(doc !== null) {
      return callback(null, doc);
    }

    request({
      url: url,
      headers: headers,
      strictSSL: strictSSL,
      followRedirect: false
    }, handleResponse);

    function handleResponse(err, res, body) {
      doc = {contextUrl: null, documentUrl: url, document: body || null};

      // handle error
      if(err) {
        return callback(new JsonLdError(
          'URL could not be dereferenced, an error occurred.',
          'jsonld.LoadDocumentError',
          {code: 'loading document failed', url: url, cause: err}), doc);
      }
      var statusText = http.STATUS_CODES[res.statusCode];
      if(res.statusCode >= 400) {
        return callback(new JsonLdError(
          'URL could not be dereferenced: ' + statusText,
          'jsonld.InvalidUrl', {
            code: 'loading document failed',
            url: url,
            httpStatusCode: res.statusCode
          }), doc);
      }

      // handle Link Header
      if(res.headers.link &&
        res.headers['content-type'] !== 'application/ld+json') {
        // only 1 related link header permitted
        var linkHeader = jsonld.parseLinkHeader(
          res.headers.link)[LINK_HEADER_REL];
        if(_isArray(linkHeader)) {
          return callback(new JsonLdError(
            'URL could not be dereferenced, it has more than one associated ' +
            'HTTP Link Header.',
            'jsonld.InvalidUrl',
            {code: 'multiple context link headers', url: url}), doc);
        }
        if(linkHeader) {
          doc.contextUrl = linkHeader.target;
        }
      }

      // handle redirect
      if(res.statusCode >= 300 && res.statusCode < 400 &&
        res.headers.location) {
        if(redirects.length === maxRedirects) {
          return callback(new JsonLdError(
            'URL could not be dereferenced; there were too many redirects.',
            'jsonld.TooManyRedirects', {
              code: 'loading document failed',
              url: url,
              httpStatusCode: res.statusCode,
              redirects: redirects
            }), doc);
        }
        if(redirects.indexOf(url) !== -1) {
          return callback(new JsonLdError(
            'URL could not be dereferenced; infinite redirection was detected.',
            'jsonld.InfiniteRedirectDetected', {
              code: 'recursive context inclusion',
              url: url,
              httpStatusCode: res.statusCode,
              redirects: redirects
            }), doc);
        }
        redirects.push(url);
        return loadDocument(res.headers.location, redirects, callback);
      }
      // cache for each redirected URL
      redirects.push(url);
      // TODO: disable cache until HTTP caching implemented
      for(var i = 0; i < redirects.length; ++i) {
        cache.set(
          redirects[i],
          {contextUrl: null, documentUrl: redirects[i], document: body});
      }
      callback(err, doc);
    }
  }
};

function buildHeaders(optionsHeaders) {
    optionsHeaders = optionsHeaders || {};

    var hasAccept = Object.keys(optionsHeaders).some(function(h) {
        return h.toLowerCase() === 'accept';
    });

    if(hasAccept) {
        throw new RangeError(
            'Accept header may not be specified as an option; only "' +
                _defaults.headers.accept + '" is supported.');
    }

    var headers = {'Accept': _defaults.headers.accept};
    for(var k in optionsHeaders) {headers[k] = optionsHeaders[k];}

    return headers;
}

jsonld.useDocumentLoader('nodeCached')

